#ifndef __relorient_approx_rot_c /* Copyright Bill.Triggs@inrialpes.fr */
#define __relorient_approx_rot_c "$ Sat Jul 22 13:01:58 2000 $"
/** NO WARRANTY. Use & distribute freely. Fixes/feedback welcome. **/
/*------------------------------------------------------------------------
  For relative orientation of two calibrated images from points,
  estimate the approximate 3D rotation between the images assuming that
  the translation is small, and also estimate the approximate modulus of
  the translation in units of the point camera distance scale.

  This is useful for detecting image pairs where the camera centre has
  hardly moved, which is a singular case for most general relative
  orientation methods: although the rotation angle can be estimated
  accurately, the translation direction and the point camera distances
  are poorly constrained owing to the small relative baseline.

  Method: Let X and Y be n x 3 matrices containing the calibrated image
  points (3D direction vectors in the camera frame), where n>=2.  We
  find the rotation matrix R that best aligns the two images, i.e that
  minimizes |Y - R*X|. If the inter-view translation t were zero, this
  would be the true inter-image rotation. To bound the error for t!=0,
  consider the standard 3D vector triangulation equation:
  
  %  zy * y = R * (zx * x) + t	  => (zy/zx) * y = R * x + t/zx

  Here, x,y are corresponding image points and zx,zy are the distances
  of the 3D point from the camera centres. Given that dy = dx + O(|t|),
  the unmodelled terms relative to y = R*x are of order O(|t|/zx), so
  for small t the error in R will also be of this order.  Moreover, if
  theta is the angle between R*x and y for the true R, we have:

  %	|sin(theta)| = |y ^ (R*x)| = |y ^ t/zx| <= |y||t|/zx = |t|/zx

  The estimated R must have a residual smaller than this, as it
  minimizes the residual. So we can turn this equation around and use
  the observed |y ^ (R*x)| values as a lower bound for |t|/zx, the
  relative camera translation in units of the camera scene distance. In
  general, provided that the y are not too tightly clustered and that
  |t|/zx is not more than about 0.1 or 0.2, this gives a reasonable
  order of magnitude estimate for |t|/zx.
  ------------------------------------------------------------------------*/
#include "geom23.h"

extern void
vecs_to_rot(int d, int n, double R[/*d x d*/], double *condz, 
	    double y[/*d x n*/], double x[/*d x n*/], int proper);

void
relorient_approx_rot(double R[3][3], double *mod_t,
		     double y[/*n*/][3], double x[/*n*/][3], int n)
{
  double cond,R1[9], mot=0.0;
  int i,j,p;

  /* Estimate rotation R.
   */
  vecs_to_rot(3,n,R1,&cond,(double*)y,(double*)x,1);
  for (i=0; i<3; i++) for (j=0; j<3; j++) R[i][j] = R1[i+3*j];

  /* Find maximum residual as estimate of |t|/z
   */
  for (p=0; p<n; p++) {
    double Rx[3], a[3],d;
    m33v3(Rx,R,x[p]);
    v3cross(a,y[p],Rx);
    d = SQ(a[X])+SQ(a[Y])+SQ(a[Z]);
    if (mot<d) mot = d;
  }
  *mod_t = mot;
}


#endif /*__relorient_approx_rot_c */
