#ifndef __vecs_to_rot_c /* Copyright Bill.Triggs@inrialpes.fr */
#define __vecs_to_rot_c "$ Fri Jul 21 15:34:43 2000 $"
/** NO WARRANTY. Use & distribute freely. Fixes/feedback welcome. **/
/*------------------------------------------------------------------------
  Estimate the rotation that maps a set of n>=d-1 d-D vectors X into
  another set Y: Y ~= R*X. Here X and Y are d x n matrices whose columns
  are the vectors, and R is a d x d rotation matrix.  

  Method: Using the invariance of trace Tr() under transpose and cyclic
  permutations, we have to find R minimizing:

  %	|Y-R*X|^2 = Tr((Y-R*X)(Y-R*X)') = Tr(Y*Y') + Tr(X*X') - 2*Tr(Y*X'*R')

  The first two terms are constant and the last is 2*Tr(S*R1') where
  U*S*V' is the SVD of Y*X' and R1 = U'*R*V. Since S is diagonal
  positive, the solution is the R1 with maximal positive weight on the
  diagonal, i.e. R1 = I, R = U*V'. Note that this method effectively
  weights the vectors according to their squared norm, so longer vectors
  have higher weight.
 ------------------------------------------------------------------------*/
#include <stdlib.h>
#include <math.h>
#include "geom23.h"

extern void
  LAPACK(dgesvd)(char jobu[], char jobv[], long* m, long* n, double A[], long* lda, 
		 double S[], double U[], long* ldu, double VT[], long* ldvt, 
		 double work[], long* lwork, long* info),
  LAPACK(dgetrf)(long* m, long* n, double A[], long* lda, long ipivot[], long* info);

void
vecs_to_rot(int d, int n, double R[/*d x d*/], double *cond, 
	    double y[/*d x n*/], double x[/*d x n*/], int proper)
{
  long dd = d, d2 = d*d, info, lwork=15;
  double *A = TMPALLOC(double,3*d2+lwork+2*d), 
    *U = A+d2, *Vt = U+d2, *S = Vt+d2, *work = S+d;
  long *ipiv = (long*)(work+lwork);
  int i,j,k;
  
  /* Form A = Y * X'.
   */
  for (i=0; i<d; i++) 
    for (j=0; j<d; j++) {
      double t = 0.0;
      for (k=0; k<n; k++) t += y[i+d*k] * x[j+d*k];
      A[i+d*j] = t;
    }
  
  /* SVD A = U * S * Vt and form R = U * Vt.
   */
  LAPACK(dgesvd)("A","A",&dd,&dd,A,&dd,S,U,&dd,Vt,&dd,work,&lwork,&info);
  
  for (i=0; i<d; i++) 
    for (j=0; j<d; j++) {
      double t = 0.0;
      for (k=0; k<d; k++) t += U[i+d*k] * Vt[k+d*j];
      R[i+d*j] = t;
    }
  
  /* Condition estimator. The method is singular iff there is at least a
   * 2D null space.  
   */
  if (cond) *cond = sqrt(SQ(S[d-2])+SQ(S[d-1])) / S[0];

  if (proper) {
    /* Force R to be proper (det(R) = 1 not -1). We find sign(det(R))
     * using LU decomposition (which gives det(R) = +-det(L*U) =
     * +-prod(diag(U)), where the sign is - iff there was an odd number
     * of row interchanges during pivoting. If the sign is negative, we
     * flip the axis corresponding to the smallest singular value, as
     * this changes Y * X' as little as possible.  
     */
    int s = 0;
    for (i=0; i<d2; i++) A[i] = R[i];
    LAPACK(dgetrf)(&dd,&dd,A,&dd,ipiv,&info);
    for (i=0; i<d; i++) {
      if (A[i+d*i]<0.0) s++;
      if (ipiv[i] != i+1) s++;
    }
    if (s % 2) {
      /* fprintf(stderr,"vecs_to_rot: flipping sign\n");*/
      for (i=0; i<d; i++) Vt[d-1+d*i] = -Vt[d-1+d*i];
      for (i=0; i<d; i++) 
	for (j=0; j<d; j++) {
	  double t = 0.0;
	  for (k=0; k<n; k++) t += U[i+d*k] * Vt[k+d*j];
	  R[i+d*j] = t;
	}
    }
  }
  TMPFREE(A);
}

#endif /*__vecs_to_rot_c */
